/* custom passwd file update program

   Copyright 1998 Malcolm Dew-Jones.  You are free to use this program
   under the terms of the GNU copyleft.
   
                 THIS PROGRAM COMES WITH NO WARRANTY.
    
    This is used to provide controlled access to custom password files.
    Three types of access are available, as compile options.  They can
    be combined.
    - update access
    - listing access
    - password verification

   Protection is provided by a three way matrix...
   1) different compiled instances of this program provide different
      access options (update, list, and password verify)
   2) file protection controls who can run which program
   3) suid or guid protection controls what passwd files each program 
      can access.

   you can compile with the following three compile #defines available
       -DLIST -DVERIFY -DUPDATE

   -------
   There is also now the ability to use flock to lock the file in case
   of shared access.  The lock is released automatically when the program
   exits.

   To compile with flock use -DFLOCK  (i.e. gcc -DFLOCK altpasswd.c)

   -------

   Lets say for example that you wish to control access to a password
   file for an application "MyApp".  Lets call the custom password file
   "MyAppPasswd.dat", and lets provide two utilities to access this file.
   One is for controlled updates by a certain group of users, and another
   allows a script to check a password is valid even though the user cannot
   read the password file.
   
   Lets compile "updateMyAppPasswd"
       gcc altpasswd -DVERIFY -DLIST -DUPDATE -o updateMyAppPasswd

   Lets compile "verifyMyAppPasswd"
       gcc altpasswd -DVERIFY -o verifyMyAppPasswd


   The intended security is as follows for update, and similar types of
   security would be setup for the other access modes.

   A _copy_ of this program is made which will be used specificaly as part of 
   the application needing the custom passwd file.

   A custom passwd file is protected via ownerships so no one can update it.
    
   The ownership of the passwd file is set to something unique, and is the 
   same as the ownership of the copy of this program.  NOTE THAT THE 
   OWNERSHIP MUST NOT BE ANYTHING LIKE ROOT, instead it should be something 
   unique which is used only for the task at hand.

   The copy of this program is then set suid or guid so it can be used to 
   update the custom password file, and the copy of the program is itself 
   protected in what ever manner is desired, so only the appropriate people 
   can run it.

   -------

    Each line of the custom password file is of the form
    username:crypted-password(13-bytes-long)[aything else]
    (We don't actually allow updates to to lines with "anything else")

    !! Notice the crypted password is a fixed length field - 
       we update it in place !!      

    - we don't handle prompting for the password yet (I only want it for 
      batch jobs anyways)
 */

/* default mode is to enable all options */
#ifndef UPDATE
#ifndef LIST
#ifndef VERIFY
#define UPDATE 1
#define LIST 1
#define VERIFY 1
#endif
#endif
#endif

#ifdef UPDATE
#define FILE_OPEN_MODE "r+"
#else
#define FILE_OPEN_MODE "r"
#endif


#define COLON_CHR ':'
#define HASH '#'
#define CRYPTED_LEN 13
#define MAX_USERNAME_LEN 24

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>

#ifdef FLOCK
#include <sys/file.h>
#ifdef FLOCK_TEST
void flock_try(void)
{
   fprintf(stderr,"flock_try() atexit(): press return\n");
   getchar();
}
#endif
#endif


#define StrEq(a,b) (strcmp((a),(b))==0)
#define StrNotEq(a,b) strcmp((a),(b))


/* From local_passwd.c (C) Regents of Univ. of California blah blah */
static unsigned char itoa64[] = /* 0 ... 63 => ascii - 64 */
"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

static void to64(register char *s, register long v, register int n)
{
    while (--n >= 0) {
    *s++ = itoa64[v & 0x3f];
    v >>= 6;
    }
}

char *encrypted(char *pw)
{
    char *cpw, salt[3];
    
    (void) srand((int) time((time_t *) NULL));
    to64(&salt[0], rand(), 2);
    salt[2] = '\0';
    cpw = (char *)crypt(pw, salt);
    return cpw;
}


static char *prognam;
void usage(void)
{
   printf(
"$RCSfile: altpasswd.c,v $ $Revision: 1.7 $\n\
Copyright 1999 Malcolm Dew-Jones under terms of GNU copyleft\n\
Usage: %s [-option(s)] file [ username ] [ password ]\n\
options: \n",prognam);
#ifdef LIST
   printf("         -l list all usernames\n");
#endif
#ifdef UPDATE
   printf("         -u update one user's password\n\
         -a add a new user and their password\n\
         -r remove a username\n\
         -C create the file if it doesn't exist (implies -a)\n");
#endif
#ifdef VERIFY
   printf("         -v validate the specified password\n");
#endif
   printf("default option: \n");
#ifdef LIST
   printf("                -l  with just a file parametre\n");
#endif
#ifdef UPDATE
#ifndef VERIFY
   printf("                -au when all three parametres are given\n");
#endif
#endif
#ifdef VERIFY
   printf("                -v  when all three parametres are given\n");
#endif
#ifdef FLOCK
   printf("This version was compiled with file locking turned on.\n");
#endif
#ifndef FLOCK
   printf("This version was NOT compiled with file locking.\n");
#endif
}

static FILE *fp=0;
interrupted(void) { fputs("*Interrupted*\n",stderr); exit(255); }


main(int argc, char *argv[])
{
   char *filename=0,*username=0, *password=0;
   char **params[3] = { &filename, &username , &password};
   int i,args;
   int optList=0,optUpdate=0,optAdd=0,optRemove=0,optCreate=0,optValidate=0;
   char EnCrypted[CRYPTED_LEN+3];

#ifdef FLOCK_TEST
atexit(flock_try);
#endif

   prognam = argv[0];
   
   signal(SIGINT,(void (*)()) interrupted );

   for (i=1,args=0; i<argc; i++)
     { if (*argv[i] == '-')
       {  char *l = argv[i]+1;
          for (;*l;l++) 
            {  switch(*l)
               { 
#ifdef LIST
                case 'l': optList=1;    break;
#endif                
#ifdef UPDATE
                case 'u': optUpdate=1; 	 break;
                case 'a': optAdd=1;    	 break;
                case 'r': optRemove=1; 	 break;
                case 'C': optCreate=1; 	 break;
#endif                
#ifdef VERIFY
                case 'v': optValidate=1; break;
#endif                
                case 'h': usage(); exit(0); 
                default: 
                  fprintf(stderr,"unknown option %c\n",*l); 
                  usage(); exit(2);
               }
            }
       }else
       {  if (args >= sizeof(params) / sizeof(params[0]))
            { fputs("Too many parametres\n",stderr); exit(2); }
          *params[args++]=argv[i];
       }
     }
   /* default options */
   if ((optList+optUpdate+optAdd+optRemove+optCreate+optValidate) == 0)
     { 
#ifdef LIST
       if (args == 1) optList=1;
       else
#endif                
#ifdef UPDATE
#ifndef VERIFY
       if (args == 3) optUpdate=optAdd=1;
       else
#endif                
#endif                
#ifdef VERIFY
       if (args == 3) optValidate=1;
       else
#endif                
       {  usage();
          exit(2);
       }
     }
     
   if (optCreate==1) optAdd = 1;

   /* ambiguous options */   
   if ( ((optList+optUpdate+optRemove+optValidate) > 1 ) ||
        ((optList+   optAdd+optRemove+optValidate) > 1 ) )
     { fputs("Which option?\n",stderr); usage(); exit(2); }

   if ( (optList   && args <  1) || 
        (optUpdate && args <  3) ||
        (optAdd    && args <  3) || 
        (optRemove && args <  2) ||
        (optValidate && args < 3)
      )
     { fputs("Too few parametres\n",stderr); exit(2); }

   if (args > 1)
   {
       if (optAdd && strpbrk(username, "\r\n:"))
       { fputs("Invalid character(s) in username\n",stderr); exit(2); 
       }
       if (strlen(username) > MAX_USERNAME_LEN) 
       {  username[MAX_USERNAME_LEN]=0;
          fprintf(stderr,"Warning: username truncated to %s\n",username);
       }
   }
      
   if (args > 2 && optValidate == 0)
     { strncpy(EnCrypted,encrypted(password),sizeof(EnCrypted)-1);
       password = EnCrypted;
     }
   
   fp = fopen(filename, FILE_OPEN_MODE );
#ifdef UPDATE
   if (optCreate && fp)
   {   fprintf(stderr,
           "File already exists: %s\nUser %s was not added to existing file\n", 
           filename,username); 
       exit(3); 
   }
   if (optCreate && !fp) 
       if (fp = fopen(filename,"w+")) printf("%s created\n",filename);
#endif
   if (!fp) 
   { fprintf(stderr,"can't open %s : ",filename); perror(0); exit(3); }

#ifdef FLOCK
#ifdef FLOCK_TEST
fprintf(stderr,"about to lock\n");
#endif
   if (flock(fileno(fp),LOCK_EX))
   { fputs("file lock failed : ",stderr); perror(0); exit(4);}
#ifdef FLOCK_TEST
fprintf(stderr,"locked, press return\n");
getchar();
#endif
#endif

#ifdef VERIFY
   if (optValidate)
     {  validate_user_password(fp,username,password);
     }else
#endif
#ifdef LIST
   if (optList)
     {  list_users(fp);
     }else
#endif
#ifdef UPDATE
   if (optUpdate)
     { update_user_password(fp,username,password,optAdd);
     }else
   if (optAdd)
     { add_user(fp,username,password);
     }else
   if (optRemove)
     { remove_user(fp,username,filename);
     }
#endif
#ifndef UPDATE
     { ; }
#endif
     
   if (fclose(fp)) 
     { fputs("Note: Error on close of file : ",stderr); perror(0); exit(4); };

   exit(0);
}
#ifdef LIST
list_users(FILE *fp)
{
    int i;
    char line[256],*colon;
    i = 0;
    while (fgets(line,sizeof(line),fp))
     {  if (line[0] == HASH) continue;
        if (!(colon = strchr(line,COLON_CHR))) continue;
        *colon = 0;
        puts(line);
       i ++;
     }
    if (i==0)
     { fprintf(stderr,"No users in file\n"); exit(1);
     }
}
#endif

static long told;
static char *colon;
static char line[256];

int find_user(FILE *fp, char *username)
{
    if (-1 == (told = ftell(fp))) 
     {fputs("ftell failed\n",stderr); perror(0); exit(4);};
    while (fgets(line,sizeof(line),fp))
     {  if (line[0] == HASH) continue;
        if (!(colon = strchr(line,COLON_CHR))) continue;
        *colon = 0;
        if (StrEq(line,username)) return 1;
        if (-1 == (told = ftell(fp))) 
          {fputs("ftell failed\n",stderr); perror(0); exit(4);};
     }
    return 0;
}
#ifdef VERIFY
validate_user_password(FILE *fp, char *username,char *pw)
{
    char salt[3];
   
    if (!find_user(fp,username)) 
    {   fprintf(stderr,"No such user: %s\n",username); exit(2);
    }

    /* force the portion of the line buffer which should have the
       old encrypted password into a nice C string */
    colon[CRYPTED_LEN+1]='\0';

    /* extract previous salt */
    salt[0] = colon[1];
    salt[1] = colon[2];
    salt[2] = '\0';

    /* encrypt the new password with the old salt and compare */
    if (strcmp((char *)crypt(pw, salt),colon+1))
    {   fprintf(stderr,"Password incorrect\n",username); exit(1);
    }
}
#endif
#ifdef UPDATE
int update_user_password(FILE *fp, char *username,char *crypted, int optAdd)
{
    if (!find_user(fp,username)) 
    if (!optAdd)
     { fprintf(stderr,"No such user: %s\n",username); exit(1);
     }else
     { /* not found, so add'em
          we're at the end of file, so just build a line and write it out */
       if (fprintf(fp,"%s%c%s\n",username,COLON_CHR,crypted) < 0)
        { fputs("file write failed : ",stderr); perror(0); exit(4);}
       printf("%s added\n",username);
       return 1;
     }
   
   /* check length (includes \n) is still the same (should be and must be) */
    if ( (strlen(colon+1)-1) != strlen(crypted))
     { fputs("bad format in file : old encrypted password wrong length",stderr);
       exit(3);
     }
    /* return to start of line */
    if (fseek(fp,told,SEEK_SET)) 
     { fputs("fseek failed to return to BOL\n",stderr); perror(0); exit(4);}
    /* overwrite old line with new line */
    if (fprintf(fp,"%s%c%s",username,COLON_CHR,crypted) < 0)
     { fputs("file write failed : ",stderr); perror(0); exit(4);}
   printf("%s updated\n",username);
   return 1;
}

int add_user(FILE *fp, char * username,char *crypted)
{   int userlen;
    if (find_user(fp,username)) 
    {   fprintf(stderr,"User already exists: %s\n",username); exit(1);
    }
    /* we're at the end of file, so build a line and write it out */
    if (fprintf(fp,"%s%c%s\n",username,COLON_CHR,crypted) < 0)
     { fputs("file write failed : ",stderr); perror(0); exit(4);}
   printf("%s added\n",username);
   return 1;
}

int remove_user(FILE *fp, char *username, char *filename)
{   
    FILE *earlier;
    if (!find_user(fp,username)) 
    {   fprintf(stderr,"No such user: %s\n",username); exit(1);
    }

    /* re open the file, so we can overlap the line to remove */
    earlier = fopen(filename,"r+");
    if (!earlier) 
     { fprintf(stderr,"can't open %s again : ",filename); perror(0); exit(4); }
    /* set the new file to the start of the line to remove */
    /* (the old file points to the next line) */
    if (fseek(earlier,told,SEEK_SET)) 
     { fputs("fseek failed to return to line\n",stderr); perror(0); exit(4);}

    /* now copy subsequent lines to earlier places in file */
    while(fgets(line,sizeof(line),fp))
     { if (EOF == fputs(line,earlier))
        { fputs("file write failed : ",stderr); perror(0); exit(4);}
     }
    if (ferror(fp))
     { fputs("file read failed : ",stderr); perror(0); exit(4);}

    /* get the current size of the re-written file */
    if (-1 == (told = ftell(earlier))) 
     {fputs("ftell failed for truncate\n",stderr); perror(0); exit(4);};

    /* and truncate the file to chop off the (repeated) last line */
    if (-1 == ftruncate(fileno(earlier),told))
     {fputs("failed to truncate file\n",stderr); perror(0); exit(4);};
     
    if (fclose(earlier))
     { fputs("Note: Error on 2nd close of file : ",stderr); perror(0); exit(4); };
    printf("%s removed\n",username);
    return 1;
}
#endif
